#include <tracker.h>
#include <stdlib.h>
#include "nautilus-metadata-extractor-tracker.h"
#include "nautilus-metadata-extractor.h"

#define METADATA_NOT_IN_TRACKER "FILE NOT IN DATABASE"

static NautilusAttributeDescriptor extension_attributes[] =
{
	{ METADATA_DURATION_ATTRIBUTE, "Duration", "Duration Description", "" },
	{ METADATA_ARTIST_ATTRIBUTE, "Artist", "Artist Description", "" },
	{ METADATA_TITLE_ATTRIBUTE, "Title", "Title Description", "" },
	{ METADATA_AUTHOR_ATTRIBUTE, "Author", "Author Description", "" },
	{ METADATA_ALBUM_ATTRIBUTE, "Album", "Album Description", "" },
	{ METADATA_YEAR_ATTRIBUTE, "Year", "Year Description", "" },
	{ METADATA_DESCRIPTION_ATTRIBUTE, "Description", "Description Description", "" },
	{ METADATA_TRACK_ATTRIBUTE, "Track No.", "Track No. Description", "" },
	{ METADATA_GENRE_ATTRIBUTE, "Genre", "Genre Description", "" },
	{ METADATA_DISC_COUNT_ATTRIBUTE, "Disc Count", "Disc Count Description", "" },
	{ METADATA_RESOLUTION_ATTRIBUTE, "Resolution", "Resolution Description", "" },
	{ METADATA_CODEC_ATTRIBUTE, "Codec", "Codec Description", "" },

	{ METADATA_AUDIO_BITRATE_ATTRIBUTE, "Bitrate", "Audio Bitrate Description", "" },
	{ METADATA_AUDIO_SAMPLE_RATE_ATTRIBUTE, "Sample Rate", "Audio Sample Rate Description", "" },
	{ METADATA_AUDIO_CHANNELS_ATTRIBUTE, "Audio Channel(s)", "Audio Channel(s) Description", "" },
	{ METADATA_AUDIO_CODEC_ATTRIBUTE, "Audio Codec", "Audio Codec Description", "" },

	{ METADATA_VIDEO_CODEC_ATTRIBUTE, "Video Codec", "Video Codec Description", "" },
	{ METADATA_VIDEO_FPS_ATTRIBUTE, "Video FPS", "Video FPS Description", "" },
	{ METADATA_VIDEO_BITRATE_ATTRIBUTE, "Video Bitrate", "Video Bitrate Description", "" },
	
	{ METADATA_DOCUMENT_SUBJECT_ATTRIBUTE, "Subject", "Subject Description", "" },
	{ METADATA_DOCUMENT_WORD_COUNT_ATTRIBUTE, "Word Count", "Word Count Description", "" },
	{ METADATA_DOCUMENT_PAGE_COUNT_ATTRIBUTE, "Page Count", "Page Count Description", "" },
	
	{ METADATA_IMAGE_CAMERA_MAKE_AND_MODEL_ATTRIBUTE, "Camera", "Camera Make and Model", "" },
};

char* extract_track(const char* value);
char* extract_resolution(const char* value);
char* extract_camera_make_model(const char* value);

typedef struct
{
	const char* tracker_attr;
	const char* metadata_attr;
} TrackerAttributeDescriptor;

static char* tracker_resolution_keys[] =
{
	"Video:Height", "Video:Width",
	"Image:Height", "Image:Width",
	NULL,
};

static char* tracker_audio_track_keys[] = 
{
	"Audio:TrackNo", "Audio:AlbumTrackCount",
	NULL,
};

static char* tracker_codec_keys[] =
{
	"Video:Codec", "Video:Bitrate", "Video:FrameRate",
	"Audio:Codec", "Audio:Bitrate", "Audio:Samplerate", "Audio:Channels",
	NULL,
};

static char* tracker_camera_make_model_attributes[] =
{
	"Image:CameraMake", "Image:CameraModel",
	NULL,
};

static TrackerAttributeDescriptor tracker_normal_string_attributes[] =
{
	{ "Audio:Title", METADATA_TITLE_ATTRIBUTE },
	{ "Audio:Artist", METADATA_ARTIST_ATTRIBUTE },
	{ "Audio:AlbumArtist", METADATA_AUTHOR_ATTRIBUTE },
	{ "Audio:Album", METADATA_ALBUM_ATTRIBUTE },
	{ "Audio:Genre", METADATA_GENRE_ATTRIBUTE },
	{ "Audio:Duration", METADATA_DURATION_ATTRIBUTE },
	{ "Audio:DiscNo", METADATA_DISC_COUNT_ATTRIBUTE },
	{ "Audio:ReleaseDate", METADATA_YEAR_ATTRIBUTE },
	{ "Audio:Comment", METADATA_DESCRIPTION_ATTRIBUTE },
	{ "Audio:Codec", METADATA_AUDIO_CODEC_ATTRIBUTE },
	{ "Audio:Samplerate", METADATA_AUDIO_SAMPLE_RATE_ATTRIBUTE },
	{ "Audio:Bitrate", METADATA_AUDIO_BITRATE_ATTRIBUTE },
	{ "Audio:Channels", METADATA_AUDIO_CHANNELS_ATTRIBUTE },
	
	{ "Video:Title", METADATA_TITLE_ATTRIBUTE },
	{ "Video:Author", METADATA_AUTHOR_ATTRIBUTE },
	{ "Video:Duration", METADATA_DURATION_ATTRIBUTE },
	{ "Video:Comments", METADATA_DESCRIPTION_ATTRIBUTE },
	{ "Video:FrameRate", METADATA_VIDEO_FPS_ATTRIBUTE },
	{ "Video:Codec", METADATA_VIDEO_CODEC_ATTRIBUTE },
	{ "Video:Bitrate", METADATA_VIDEO_BITRATE_ATTRIBUTE },
	
	{ "Image:Title", METADATA_TITLE_ATTRIBUTE },
	{ "Image:Album", METADATA_ALBUM_ATTRIBUTE },
	{ "Image:Date", METADATA_DATE_ATTRIBUTE },
	{ "Image:Creator", METADATA_AUTHOR_ATTRIBUTE },
	{ "Image:Comments", METADATA_DESCRIPTION_ATTRIBUTE },
	
	{ "Doc:Title",  METADATA_TITLE_ATTRIBUTE },
	{ "Doc:Subject", METADATA_DOCUMENT_SUBJECT_ATTRIBUTE },
	{ "Doc:Author",  METADATA_AUTHOR_ATTRIBUTE },
	{ "Doc:Comments",  METADATA_DESCRIPTION_ATTRIBUTE },
	{ "Doc:PageCount", METADATA_DOCUMENT_PAGE_COUNT_ATTRIBUTE },
	{ "Doc:WordCount", METADATA_DOCUMENT_WORD_COUNT_ATTRIBUTE },
	{ "Doc:Created", METADATA_DATE_ATTRIBUTE },
};

static ServiceType service;

// this function is used to set the track count tag in the format (one of the followings)
// track_num/total_track_num
// track_num
// ?/total_track_num
void set_track_count(Metadata* metadata, char** values)
{
	char* audio_track = NULL;
	gboolean found = FALSE;
	
	if (! g_str_equal(values[0], ""))
	{
		found = TRUE;
		if (g_str_equal(values[1], ""))
			audio_track = g_strdup_printf("%s", values[0]);
		else
			audio_track = g_strdup_printf("%s/%s", values[0], values[1]);
	}
	else if (! g_str_equal(values[1], ""))
	{
		found = TRUE;
		audio_track = g_strdup_printf("?/%s", values[1]);
	}
	
	
	if (found)
	{
		metadata_set(metadata, METADATA_TRACK_ATTRIBUTE, audio_track);
		g_free(audio_track);
	}
}

// this function is used to set the resollution tag in the format (one of the followings)
// res_x x res_y
// ? x res_y
// res_x x ?
void set_resolution(Metadata* metadata, char** values)
{
	char* resolution;
	const char *x="?", *y="?";
	gboolean found = FALSE;

	if (! g_str_equal(values[0], ""))
	{
		y = values[0];
		found = TRUE;
	}
	if (! g_str_equal(values[1], ""))
	{
		x = values[1];
		found = TRUE;
	}

	if (found)
	{
		resolution = g_strdup_printf("%s x %s", x, y);
		metadata_set(metadata, METADATA_TRACK_ATTRIBUTE, resolution);
		g_free(resolution);
	}
}

// this function will merge the fotocamera make and camera model in one single tag
void set_camera_make_and_model(Metadata* metadata, char** values)
{
	if (! (g_str_equal(values[0], "") && g_str_equal(values[1], "")) )
	{
		char* make_model = g_strdup_printf("%s %s", values[0], values[1]);
		metadata_set(metadata, METADATA_IMAGE_CAMERA_MAKE_AND_MODEL_ATTRIBUTE, make_model);
		g_free(make_model);
	}
}

// this function will merge the audio and video codec, bitrate, samplerate, framerate, channels.
void set_codec(Metadata* metadata, char** values)
{
	char* codec = NULL;
	char* audio_codec = NULL;
	char* video_codec = NULL;
	
	if (!g_str_equal(values[3], ""))
	{
		int bitrate = atoi(values[4]);
		bitrate /= 1000; // or maybe 1024?? don't know
		int samplerate = atoi(values[5]);
		samplerate /= 1000;  // or maybe 1024?? don't know
		audio_codec = g_strdup_printf("%s %dkb/s %dkhz %s", values[3], bitrate, samplerate, values[6]);
	}
	
	if (!g_str_equal(values[0], ""))
	{
		int bitrate = atoi(values[1]);
		bitrate /= 1000; // or maybe 1024?? don't know
		video_codec = g_strdup_printf("%s %dkb/s %sfps", values[0], bitrate, values[2]);
		
		if (audio_codec != NULL)
		{
			codec = g_strdup_printf("%s (Audio: %s)", video_codec, audio_codec);
		}
		else
		{
			codec = g_strdup(video_codec);
		}
	}
	
	if (audio_codec != NULL)
		g_free(audio_codec);
	
	if (video_codec != NULL)
		g_free(video_codec);
	
	if (codec != NULL)
	{
		metadata_set(metadata, METADATA_CODEC_ATTRIBUTE, codec);
		g_free(codec);
	}
}

// this function must extract metadata from file_uri and insert it into the Metadata struct
void extract_metadata_file(Metadata* metadata, const char* file_uri, const char* mime_type)
{
	TrackerClient* client = tracker_connect(FALSE);
	
	// get the path of the current file: tracker unfortunately only supports path and not uris
	GFile* file = g_file_new_for_uri(file_uri);
	char* path = g_file_get_path(file);
	g_object_unref(G_OBJECT(file));
	
	char** values = NULL;
	// create the keys to pass to tracker. In this step only normal keys are used
	// normal keys are keys which will return a value that can be used as it is:
	// for example Artist, but not resolution, since the resolution is divided into
	// Height and Width, but it's better to merge them
	char** keys = g_new(char*, G_N_ELEMENTS(tracker_normal_string_attributes)+1);
	
	// move outside as a static variable!
	int i;
	for (i=0; i<G_N_ELEMENTS(tracker_normal_string_attributes); i++)
	{
		keys[i] = (char*) tracker_normal_string_attributes[i].tracker_attr;
	}
	keys[i] = NULL;
	
	values = tracker_metadata_get(client, service, path, keys, NULL);
	
	if (values != NULL)
	{
		int i=0;
		while (values[i] != NULL)
		{
			const char *metadata_attribute_name = tracker_normal_string_attributes[i].metadata_attr;
			char *attribute_value = values[i];
			
			if (attribute_value != NULL && !g_str_equal(attribute_value, ""))
				metadata_set(metadata, metadata_attribute_name, attribute_value);
			
			g_free(values[i]);
			i++;
		}
		g_free(values);
	}
	else
	{
		// this directory is not watched by tracker, inform the user
		for (i=0; i<G_N_ELEMENTS(tracker_normal_string_attributes); i++)
			metadata_set(metadata, tracker_normal_string_attributes[i].metadata_attr, METADATA_NOT_IN_TRACKER);
	}
	g_free(keys);
	
	// under this line special couple of keys, whose value result must be parsed to form a unique Metadata tag, are considered
	// track numbers in form of "tr_num/tr_count"

	values = tracker_metadata_get(client, service, path, tracker_audio_track_keys, NULL);
	if (values != NULL)
	{
		set_track_count(metadata, values);
		g_free(values[0]);
		g_free(values[1]);
		g_free(values);
	}
	else
	{
		// this directory is not watched by tracker, inform the user
		metadata_set(metadata, METADATA_TRACK_ATTRIBUTE, METADATA_NOT_IN_TRACKER);
	}
	
	// resolution in the form of "width x height"
	values = tracker_metadata_get(client, service, path, tracker_resolution_keys, NULL);
	if (values != NULL)
	{
		int i=0;
		while (values[i] != NULL)
		{
			set_resolution(metadata, values + i);
			g_free(values[i]);
			i++;
			g_free(values[i]);
			i++;
		}
		g_free(values);
	}
	else
	{
		// this directory is not watched by tracker, inform the user
		metadata_set(metadata, METADATA_RESOLUTION_ATTRIBUTE, METADATA_NOT_IN_TRACKER);
	}
	
	// fotocamera make and model
	values = tracker_metadata_get(client, service, path, tracker_camera_make_model_attributes, NULL);
	if (values != NULL)
	{
		set_camera_make_and_model(metadata, values);
		g_free(values[0]);
		g_free(values[1]);
		g_free(values);
	}
	else
	{
		// this directory is not watched by tracker, inform the user
		metadata_set(metadata, METADATA_IMAGE_CAMERA_MAKE_AND_MODEL_ATTRIBUTE, METADATA_NOT_IN_TRACKER);
	}
	
	values = tracker_metadata_get(client, service, path, tracker_codec_keys, NULL);
	if (values != NULL)
	{
		set_codec(metadata, values);
		g_free(values[0]);
		g_free(values[1]);
		g_free(values[2]);
		g_free(values[3]);
		g_free(values[4]);
		g_free(values[5]);
		g_free(values[6]);
		g_free(values);
	}
	else
	{
		// this directory is not watched by tracker, inform the user
		metadata_set(metadata, METADATA_CODEC_ATTRIBUTE, METADATA_NOT_IN_TRACKER);
	}

	tracker_disconnect(client);
	g_free(path);
}

// this function must extract metadata from directory_uri and insert it into the Metadata struct
// it is not a problem if the body of this function blank (directories hardly ever have metadata)
void extract_metadata_directory(Metadata* metadata, const char* directory_uri)
{
	// This will be useful in the future if nautilus will let change the metadata directory at runtime
	// to force the list view in video/music/pictures directories
}

// returns: the maximun number of contemporary running threads
// a value < 1 is ignored
// each thread executes either extract_metadata_file or extract_metadata_directory
int get_max_thread_number(void)
{
	return 0;
}

// this function is called when the extension is registered to nautilus
void initialize_extractor(void)
{
}

// this function is called whene the extension is un-registered from nautilus
void shutdown_extractor(void)
{
	service = tracker_service_name_to_type("*");
}

// returns a list of NautilusAttributeDescriptor that must be filled with the attributes that
// the extractor can handle
// the caller will call g_list_free() so you must not store the list outside the function
GList* get_supported_attributes()
{
	GList* supported_attributes = NULL;
	
	int i;
	for (i=0; i<G_N_ELEMENTS(extension_attributes); i++)
	{
		supported_attributes = g_list_prepend(supported_attributes, extension_attributes + i);
	}
	
	return supported_attributes;
}

// this function must return the extension name, that must be unique in all nautilus-
// an example "NautilusMetadataExtensionDummy"
const char* get_extension_name()
{
	return "NautilusMetadataFileInfoTracker";
}