#include <libnautilus-extension/nautilus-info-provider.h>
#include "nautilus-metadata-info.h"
#include "nautilus-metadata-extractor.h"

#define FILE_HAS_METADATA "metadata::has-metadata"

static GMutex* mutex;
static GCond* cond;
static guint current_running_threads;
static guint current_waiting_threads;

static GList* supported_attributes;

void metadata_foreach_tag(GQuark key_id, char* value, NautilusFileInfo* file)
{
    NautilusFileInfoIface* iface = NAUTILUS_FILE_INFO_GET_IFACE(file);
    iface->add_string_attribute(file, g_quark_to_string(key_id), value);
}

Metadata* metadata_new()
{
	Metadata* metadata = g_new(Metadata, 1);
	metadata->gdata = g_new(GData*, 1);
	g_datalist_init(metadata->gdata);
/*
	metadata->mutex = g_mutex_new();
*/
	
	GList* attributes = supported_attributes;
	while (attributes != NULL)
	{
		NautilusAttributeDescriptor* attribute_descriptor = (NautilusAttributeDescriptor*) attributes->data;
		const char* attribute_name = attribute_descriptor->attribute;
		const char* attribute_default_value = attribute_descriptor->default_value;
		if (attribute_default_value == NULL)
			attribute_default_value = "";
		metadata_set(metadata, attribute_name, attribute_default_value);
		attributes = g_list_next(attributes);
	}
	
	return metadata;
}

void metadata_free(Metadata* datalist)
{
	g_return_if_fail(datalist != NULL);
	
/*
	g_mutex_free(datalist->mutex);
*/
	g_datalist_clear(datalist->gdata);
	g_free(datalist->gdata);
	g_free(datalist);
}

void metadata_set(Metadata* metadata, const char* attribute_name, const char* attribute_value)
{
	g_return_if_fail(metadata != NULL);
	g_return_if_fail(attribute_name != NULL);
	g_return_if_fail(attribute_value != NULL);
	
/*
	g_mutex_lock(metadata->mutex);	
*/
	g_datalist_id_set_data_full(metadata->gdata, g_quark_from_string(attribute_name), g_strdup(attribute_value), g_free);
/*
	g_mutex_unlock(metadata->mutex);
*/
}

const char* metadata_get(Metadata* metadata, const char* attribute_name)
{
	g_return_val_if_fail(metadata != NULL, NULL);
	g_return_val_if_fail(attribute_name != NULL, NULL);
	
	char* ret;
/*
	g_mutex_lock(metadata->mutex);
*/
	ret = g_datalist_id_get_data(metadata->gdata, g_quark_from_string(attribute_name));
/*
	g_mutex_unlock(metadata->mutex);
*/
	return ret;
}

void metadata_foreach(Metadata* metadata, MetadataForeachFunc func, gpointer user_data)
{
	g_return_if_fail(metadata != NULL);
	
/*
	g_mutex_lock(metadata->mutex);
*/
	g_datalist_foreach(metadata->gdata, func, user_data);
/*
	g_mutex_unlock(metadata->mutex);
*/
}

gpointer fill_metadata_async(UpdateHandle* handle)
{
	int max_threads = get_max_thread_number();
	// THREAD LIMIT
	if (max_threads > 1)
	{
		g_mutex_lock(mutex);
		if (current_running_threads >= max_threads)
		{
			current_waiting_threads++;
			g_cond_wait(cond, mutex);
		}
		current_running_threads++;
		g_mutex_unlock(mutex);
	}
	// THREAD LIMIT END
	
	NautilusFileInfo* file = handle->file;
	NautilusFileInfoIface* file_iface = NAUTILUS_FILE_INFO_GET_IFACE(file);
	
	Metadata* metadata = NULL;
	if (! handle->cancelled)
	{
		char* uri = file_iface->get_uri(file);
		metadata = metadata_new();
		
		if (file_iface->is_directory(file))
		{
			extract_metadata_directory(metadata, uri);
		}
		else
		{
			char* mime_type = file_iface->get_mime_type(file);
			extract_metadata_file(metadata, uri, mime_type);
			g_free(mime_type);
		}
		
		g_free(uri);
	}
		
	if (metadata != NULL && ! handle->cancelled)
	{
		metadata_foreach(metadata, (GDataForeachFunc) metadata_foreach_tag, file);
		
		// metadata caching
		g_object_set_data_full (G_OBJECT (handle->file), FILE_HAS_METADATA, metadata, (GDestroyNotify) metadata_free);
	}
	else if (metadata != NULL)
	{
		metadata_free(metadata);
	}
	
	if (! handle->cancelled)
	{
		// unlock nautilus from waiting for us
		nautilus_info_provider_update_complete_invoke(handle->update_complete, handle->provider, (NautilusOperationHandle*)handle, NAUTILUS_OPERATION_COMPLETE);
	}
	
	g_closure_unref (handle->update_complete);
	g_object_unref (handle->file);
	g_free (handle);
	
	// WAKE UP OTHER THREADS
	if (max_threads > 1)
	{
		g_mutex_lock(mutex);
		current_running_threads--;
		if (current_waiting_threads > 0)
		{
			g_cond_signal(cond);
			current_waiting_threads--;
		}
		g_mutex_unlock(mutex);
	}
	
	// GThreadFunc requires a return value
	return NULL;
}

NautilusOperationResult metadata_info_update_file_info(NautilusInfoProvider *provider, NautilusFileInfo *file, GClosure *update_complete, NautilusOperationHandle **handle)
{
	Metadata* metadata = g_object_get_data (G_OBJECT (file), FILE_HAS_METADATA);
	if (metadata != NULL)
	{
		/* FILL NAUTILUS ATTRIBUTES USING CACHED RESULTS */
		metadata_foreach(metadata, (MetadataForeachFunc) metadata_foreach_tag, file);
		return NAUTILUS_OPERATION_COMPLETE;
	}
	else
	{
		/* NO CACHED RESTULTS FOUND */
		UpdateHandle *update_handle = g_new (UpdateHandle, 1);
		update_handle->update_complete = g_closure_ref (update_complete);
		update_handle->provider = provider;
		update_handle->file = g_object_ref (file);
		update_handle->cancelled = FALSE;
		*handle = (NautilusOperationHandle*) update_handle;
		
		/* ASYNCHRONOUSLY FILL NAUTILUS ATTRIBUTES USING FUNCTION async(update_handle) */
		if (! g_thread_create((GThreadFunc)fill_metadata_async, update_handle, FALSE, NULL))
		{
			return NAUTILUS_OPERATION_FAILED;
		}
	
		return NAUTILUS_OPERATION_IN_PROGRESS;
	}
}

void initialize(void)
{
	mutex = g_mutex_new();
	cond = g_cond_new();
	supported_attributes = get_supported_attributes();
	initialize_extractor();
}

void shutdown(void)
{
	g_mutex_free(mutex);
	g_cond_free(cond);
	g_list_free(supported_attributes);
	shutdown_extractor();
}